//-----------------------------------------------------------------------------
// Project            	GeometryKit
// Class		G3DQuaterniond
// Creator            	Philippe C.D. Robert
// Maintainer         	Philippe C.D. Robert
// Creation Date      	Thu Sep  9 11:05:08 CEST 1999
//
// Copyright (c) Philippe C.D. Robert
//
// The GeometryKit is free software; you can redistribute it and/or modify it 
// under the terms of the GNU LGPL Version 2 as published by the Free 
// Software Foundation
//
// $Id: G3DQuaterniond.m,v 1.1 2002/10/12 17:24:32 probert Exp $
//-----------------------------------------------------------------------------

#import "G3DQuaterniond.h"

#import "G3DMatrix3d.h"
#import "G3DMatrix4d.h"
#import "G3DVector3d.h"
#import "G3DTuple4d.h"

#import "G3DDefs.h"
#import "G3DQuatFunc.h"
#import "G3DVectorFunc.h"
#import "G3DMatrixFunc.h"

NSString *G3DQuaterniondException = @"G3DQuaterniondException";

@implementation G3DQuaterniond

//-----------------------------------------------------------------------------
// class methods
//-----------------------------------------------------------------------------

static Class G3DTuple4dClass;
static Class G3DMatrix4dClass;
static Class G3DMatrix3dClass;

+ (void)initialize
{
  if (self == [G3DQuaterniond class]) {
    G3DTuple4dClass  = [G3DTuple4d class];
    G3DMatrix4dClass  = [G3DMatrix4d class];
    G3DMatrix3dClass  = [G3DMatrix3d class];
  }
}

//-----------------------------------------------------------------------------
// init and free
//-----------------------------------------------------------------------------

- (id)init
{
  if ((self = [super initWithX:0.0 y:0.0 z:0.0 w:1.0])) {
  }
  return self;
}

- (id)initWithElements:(double *)values
{
  if ((self = [self init])) {
    G3DCopyVector4dv(tuple,values);
  }
  return self;
}

- (id)initWithQuaternion:(G3DQuaterniond *)aTuple
{
  if ((self = [self init])) {
    G3DCopyVector4dv(tuple,[aTuple elements]);
  }
  return self;
}

- (id)initWithMatrix4d:(G3DMatrix4d *)aMatrix
{
  if ((self = [self init])) {
    G3DQuatFromMatrixd(tuple,[aMatrix elements]);
  }
  return self;
}

- (id)initWithEulerRep:(double *)ev
{
  if ((self = [self init])) {
    G3DQuatFromEulerRepd(tuple,ev);
  }
  return self;
}

//-----------------------------------------------------------------------------
// Math
//-----------------------------------------------------------------------------

- (double)norm
{
  return (tuple[0]*tuple[0] + 
          tuple[1]*tuple[1] + 
          tuple[2]*tuple[2] + 
          tuple[3]*tuple[3]);
}

- (void)conjugate
{
  tuple[0] = -tuple[0];
  tuple[1] = -tuple[1];
  tuple[2] = -tuple[2];
}

- (G3DQuaterniond *)conjugatedQuaternion
{
  G3DQuaterniond *tmp = [[G3DQuaterniond alloc] initWithQuaternion:self];

  [tmp conjugate];
  
  return [tmp autorelease];
}

- (void)invert
{
  register double sql = tuple[0]*tuple[0] + 
                        tuple[1]*tuple[1] + 
                        tuple[2]*tuple[2] + 
                        tuple[3]*tuple[3];

  if (sql != 0.0) {
    sql = 1.0/sql;
  }
  else {
    sql = 1.0;
  }

  tuple[0] *= -sql;
  tuple[1] *= -sql;
  tuple[2] *= -sql;
  tuple[3] *=  sql;
}

- (G3DQuaterniond *)invertedQuaternion
{
  G3DQuaterniond *tmp = [[G3DQuaterniond alloc] initWithQuaternion:self];

  [tmp invert];
  return [tmp autorelease];
}

- (void)normalise
{
  register double sql = tuple[0]*tuple[0] + 
                        tuple[1]*tuple[1] + 
                        tuple[2]*tuple[2] + 
                        tuple[3]*tuple[3];
  
  if (sql > 0.0) {
    sql = (1.0/ sqrt(sql));

    tuple[0] *= sql;
    tuple[1] *= sql;
    tuple[2] *= sql;
    tuple[3] *= sql;
  }
}

- (G3DQuaterniond *)normalisedQuaternion
{
  G3DQuaterniond *tmp = [[G3DQuaterniond alloc] initWithQuaternion:self];

  [tmp normalise];
  return [tmp autorelease];
}

- (void)interpolate:(G3DQuaterniond *)quat factor:(double)val
{
  double tmp[4];
  double *to = [quat elements];

  G3DInterpolateQuatd(tmp,tuple,to,val);
  G3DCopyVector4dv(tuple,tmp);
}

- (void)multiplyByQuaternion:(G3DQuaterniond *)aQuat
{
  double *a = [aQuat elements];

  G3DMultiplyQuatd(tuple, tuple, a);
}

- (void)multiplyByQuaternion:(G3DQuaterniond *)quatA and:(G3DQuaterniond *)quatB
{
  double *a = [quatA elements];
  double *b = [quatB elements];

  G3DMultiplyQuatd(tuple, a, b);
}

- (void)multiplyByInvertedQuaternion:(G3DQuaterniond *)aQuat
{
  [self multiplyByQuaternion:[aQuat invertedQuaternion]];
}

- (void)multiplyByQuaternion:(G3DQuaterniond *)a andInverted:(G3DQuaterniond *)b
{
  [self multiplyByQuaternion:a and:[b invertedQuaternion]];
}

- (void)multiplyByScalar:(double)value
{
  tuple[0] *= value;
  tuple[1] *= value;
  tuple[2] *= value;
  tuple[3] *= value;
}

- (void)addQuaternion:(G3DQuaterniond *)aQuat
{
  double *t = [aQuat elements];

  tuple[0] += t[0];
  tuple[1] += t[1];
  tuple[2] += t[2];
  tuple[3] += t[3];
}

- (void)subQuaternion:(G3DQuaterniond *)aQuat
{
  double *t = [aQuat elements];

  tuple[0] -= t[0];
  tuple[1] -= t[1];
  tuple[2] -= t[2];
  tuple[3] -= t[3];
}

- (void)rotateByAngle:(double)angle axis:(G3DVector3d *)aVec
{
  double tmp[4];
  double *v = [aVec elements];
  double *a = tuple;
  double *b = tmp;

  G3DCopyVector4d(tmp,v[0],v[1],v[2],angle);
  G3DMultiplyQuatd(tuple, a, b);

  [self normalise];
}

//-----------------------------------------------------------------------------
// Accessor methods
//-----------------------------------------------------------------------------

- (void)setQuaternionWithMatrix4d:(G3DMatrix4d *)m
{
  double *mat = [m elements];

  G3DQuatFromMatrixd(tuple,mat);
}

- (G3DMatrix4d *)rotationMatrix4d
{
  double m[16];

  G3DMatrixFromQuatd(m,tuple);

  return [[[G3DMatrix4dClass alloc] initWithElements:m] autorelease];
}

- (G3DMatrix3d *)rotationMatrix3d
{
  double m[9];

  double x2x = tuple[0] * (tuple[0] + tuple[0]);
  double x2y = tuple[0] * (tuple[1] + tuple[1]);
  double x2z = tuple[0] * (tuple[2] + tuple[2]);

  double w2x = tuple[3] * (tuple[0] + tuple[0]);
  double w2y = tuple[3] * (tuple[1] + tuple[1]);
  double w2z = tuple[3] * (tuple[2] + tuple[2]);

  double y2y = tuple[1] * (tuple[1] + tuple[1]);
  double y2z = tuple[1] * (tuple[2] + tuple[2]);
  double z2z = tuple[2] * (tuple[2] + tuple[2]);

  G3DCopyVector3d(m, 1.0-(y2y+z2z), x2y-w2z, x2z+w2y);
  G3DCopyVector3d(m+3, x2y+w2z, 1.0-(x2x+z2z), y2z-w2x);
  G3DCopyVector3d(m+6, x2z-w2y, y2z+w2x, 1.0-(x2x+y2y));

  return [[[G3DMatrix3dClass alloc] initWithElements:m] autorelease];
}

- (void)setQuaternion:(G3DQuaterniond *)q
{
  [super setValuesWithTuple:q];
}

- (void)setVector:(G3DVector3d *)vec angle:(double)val
{
  double *tmp = [vec elements];

  G3DQuatFromAngleAxisd(tuple,val,tmp);
}

- (G3DTuple4d *)angleAxisRepresentation
{
  double angleAxis[4];

  G3DAngleAxisFromQuatd(angleAxis,angleAxis,tuple);

  return [[[G3DTuple4dClass alloc] initWithElements:angleAxis] autorelease];
}

@end






