//-----------------------------------------------------------------------------
// Project            	RenderKit
// Class		G3DTuple2f
// Creator            	Philippe C.D. Robert
// Maintainer         	Philippe C.D. Robert
// Creation Date      	Thu Sep  9 11:05:08 CEST 1999
//
// Copyright (c) Philippe C.D. Robert
//
// The GeometryKit is free software; you can redistribute it and/or modify it 
// under the terms of the GNU LGPL Version 2 as published by the Free 
// Software Foundation
//
// $Id: G3DTuple2f.m,v 1.1 2002/10/12 17:24:32 probert Exp $
//-----------------------------------------------------------------------------

#import "G3DTuple2f.h"
#import "G3DFunctions.h"
#import "G3DDefs.h"
#import "G3DVectorFunc.h"
#import "G3DMatrixFunc.h"

NSString *G3DTuple2fException = @"G3DTuple2fException";

@implementation G3DTuple2f

//-----------------------------------------------------------------------------
// Class methods
//-----------------------------------------------------------------------------

static Class NSArrayClass;
static Class NSMutableArrayClass;
static Class NSStringClass;

+ (void)initialize
{
  if (self == [G3DTuple2f class]) {
    NSStringClass = [NSString class];
    NSArrayClass  = [NSArray class];
    NSMutableArrayClass  = [NSMutableArray class];
  }
}

//-----------------------------------------------------------------------------
// init and free
//-----------------------------------------------------------------------------

- (id)init
{
  return [self initWithX:0.0 y:0.0];
}

- (id)initWithElements:(float *)vals
{
  return [self initWithX:vals[0] y:vals[1]];
}

- (id)initWithX:(float)x y:(float)y
{
  if ((self = [super init])) {
    tuple[0] = x;
    tuple[1] = y;
  }
  return self;
}

- (id)initWithTuple:(G3DTuple2f *)aTuple
{
  return [self initWithX:[aTuple x] y:[aTuple y]];
}

//-----------------------------------------------------------------------------
// Math
//-----------------------------------------------------------------------------

- (void)absolute
{
  tuple[0] = ABS(tuple[0]);
  tuple[1] = ABS(tuple[1]);
}

- (void)clamp
{
  tuple[0] = CLAMP(tuple[0],0.0,1.0);
  tuple[1] = CLAMP(tuple[1],0.0,1.0);
}

- (void)clampLow:(float)low high:(float)high
{
  if (low > high) {
    float tmp = low;
    low = high;
    high = tmp;
  }
  tuple[0] = CLAMP(tuple[0],low,high);
  tuple[1] = CLAMP(tuple[1],low,high);
}

- (void)addTuple2f:(G3DTuple2f *)aTuple
{
  float *tmp = [aTuple elements];

  tuple[0] += tmp[0];
  tuple[1] += tmp[1];
}

- (void)subTuple2f:(G3DTuple2f *)aTuple
{
  float *tmp = [aTuple elements];

  tuple[0] -= tmp[0];
  tuple[1] -= tmp[1];
}

- (void)multiplyBy:(float)aScalar
{
  tuple[0] *= aScalar;
  tuple[1] *= aScalar;
}

- (void)divideBy:(float)aScalar
{
  if (aScalar == 0.0) {
    [NSException raise:G3DTuple2fException format:@"Division by zero!"];
  }
  
  tuple[0] /= aScalar;
  tuple[1] /= aScalar;
}

- (void)interpolateBetween:(G3DTuple2f *)first and:(G3DTuple2f *)second factor:(float)factor
{
  float ifac = 1 - factor;
  float *a = [first elements];
  float *b = [second elements];

  tuple[0] = ifac*a[0] + factor*b[0];
  tuple[1] = ifac*a[0] + factor*b[0];
}

- (void)negate
{
  tuple[0] *= -1.0;
  tuple[1] *= -1.0;
}

- (BOOL)isEqualToTuple:(id)aTuple
{
  if ([aTuple isKindOfClass:[G3DTuple2f class]]) {
    if (tuple[0] == [aTuple x] && tuple[1] == [aTuple y]) {
      return YES;
    }
  }
  return NO;
}

//-----------------------------------------------------------------------------
// Accessing
//-----------------------------------------------------------------------------

- (float *)elements
{
  return tuple;
}

- (void)setElements:(float *)values
{
  G3DCopyVector2fv(tuple,values);
}

- (void)getElements:(float *)values
{
  G3DCopyVector2fv(values,tuple);
}

- (float)x
{
  return tuple[0];
}

- (void)setX:(float)x
{
  tuple[0] = x;
}

- (float)y
{
  return tuple[1];
}

- (void)setY:(float)y
{
  tuple[1] = y;
}

- (void)setValuesWithTuple:(G3DTuple2f *)aTuple
{
  float *tmp = [aTuple elements];
 
  tuple[0] = tmp[0];
  tuple[1] = tmp[1];
}

- (NSString *)description
{
  return [NSStringClass stringWithFormat:@"<%@ %x %x>: x = %f, y = %f",[self class],self,[NSThread currentThread],tuple[0],tuple[1]];
}

//-----------------------------------------------------------------------------
// NSCoding
//-----------------------------------------------------------------------------

- (void)encodeWithCoder:(NSCoder *)aCoder
{
  [aCoder encodeArrayOfObjCType:@encode(float) count:2 at:tuple];
}

- (id)initWithCoder:(NSCoder *)aCoder
{
  self = [super init];
  [aCoder decodeArrayOfObjCType:@encode(float) count:2 at:tuple];
  return self;
}

//-----------------------------------------------------------------------------
// NSCopying
//-----------------------------------------------------------------------------

- (id)copyWithZone:(NSZone *)zone
{
  return [[[self class] allocWithZone:zone] initWithTuple:self];
}

@end






