//-----------------------------------------------------------------------------
// Project            	RenderKit
// Class		G3DTuple4d
// Creator            	Philippe C.D. Robert
// Maintainer         	Philippe C.D. Robert
// Creation Date      	Thu Sep  9 11:05:08 CEST 1999
//
// Copyright (c) Philippe C.D. Robert
//
// The GeometryKit is free software; you can redistribute it and/or modify it 
// under the terms of the GNU LGPL Version 2 as published by the Free 
// Software Foundation
//
// $Id: G3DTuple4d.m,v 1.1 2002/10/12 17:24:32 probert Exp $
//-----------------------------------------------------------------------------

#import "G3DTuple4d.h"
#import "G3DFunctions.h"
#import "G3DDefs.h"
#import "G3DVectorFunc.h"
#import "G3DMatrixFunc.h"

NSString *G3DTuple4dException = @"G3DTuple4dException";

@implementation G3DTuple4d

//-----------------------------------------------------------------------------
// Class methods
//-----------------------------------------------------------------------------

static Class NSArrayClass;
static Class NSMutableArrayClass;
static Class NSStringClass;

+ (void)initialize
{
  if (self == [G3DTuple4d class]) {
    NSStringClass = [NSString class];
    NSArrayClass  = [NSArray class];
    NSMutableArrayClass  = [NSMutableArray class];
  }
}

//-----------------------------------------------------------------------------
// init and free
//-----------------------------------------------------------------------------

- (id)init
{
  return [self initWithX:0.0 y:0.0 z:0.0 w:0.0];
}

- (id)initWithElements:(double *)vals
{
  return [self initWithX:vals[0] y:vals[1] z:vals[2] w:vals[3]];
}

- (id)initWithX:(double)x y:(double)y z:(double)z w:(double)w
{
  if ((self = [super init])) {
    tuple[0] = x;
    tuple[1] = y;
    tuple[2] = z;
    tuple[3] = w;
  }
  return self;
}

- (id)initWithTuple:(G3DTuple4d *)aTuple
{
  return [self initWithX:[aTuple x] y:[aTuple y] z:[aTuple z] w:[aTuple w]];
}

//-----------------------------------------------------------------------------
// Math
//-----------------------------------------------------------------------------

- (void)absolute
{
  tuple[0] = ABS(tuple[0]);
  tuple[1] = ABS(tuple[1]);
  tuple[2] = ABS(tuple[2]);
  tuple[3] = ABS(tuple[3]);
}

- (void)clamp
{
  tuple[0] = CLAMP(tuple[0],0.0,1.0);
  tuple[1] = CLAMP(tuple[1],0.0,1.0);
  tuple[2] = CLAMP(tuple[2],0.0,1.0);
  tuple[3] = CLAMP(tuple[3],0.0,1.0);
}

- (void)clampLow:(double)low high:(double)high
{
  if (low > high) {
    double tmp = low;
    low = high;
    high = tmp;
  }
  tuple[0] = CLAMP(tuple[0],low,high);
  tuple[1] = CLAMP(tuple[1],low,high);
  tuple[2] = CLAMP(tuple[2],low,high);
  tuple[3] = CLAMP(tuple[3],low,high);
}

- (void)addTuple4d:(G3DTuple4d *)aTuple
{
  double *tmp = [aTuple elements];

  tuple[0] += tmp[0];
  tuple[1] += tmp[1];
  tuple[2] += tmp[2];
  tuple[3] += tmp[3];
}

- (void)subTuple4d:(G3DTuple4d *)aTuple
{
  double *tmp = [aTuple elements];

  tuple[0] -= tmp[0];
  tuple[1] -= tmp[1];
  tuple[2] -= tmp[2];
  tuple[3] -= tmp[3];
}

- (void)multiplyBy:(double)aScalar
{
  tuple[0] *= aScalar;
  tuple[1] *= aScalar;
  tuple[2] *= aScalar;
  tuple[3] *= aScalar;
}

- (void)divideBy:(double)aScalar
{
  if (aScalar == 0.0) {
    [NSException raise:G3DTuple4dException format:@"Division by zero!"];
  }
  
  tuple[0] /= aScalar;
  tuple[1] /= aScalar;
  tuple[2] /= aScalar;
  tuple[3] /= aScalar;
}

- (void)interpolateBetween:(G3DTuple4d *)first and:(G3DTuple4d *)second factor:(double)factor
{
  double ifac = 1 - factor;
  double *a = [first elements];
  double *b = [second elements];

  tuple[0] = ifac*a[0] + factor*b[0];
  tuple[1] = ifac*a[0] + factor*b[0];
  tuple[2] = ifac*a[0] + factor*b[0];
}

- (void)negate
{
  tuple[0] *= -1.0;
  tuple[1] *= -1.0;
  tuple[2] *= -1.0;
  tuple[3] *= -1.0;
}

- (BOOL)isEqualToTuple:(id)aTuple
{
  if ([aTuple isKindOfClass:[G3DTuple4d class]]) {
    if (tuple[0] == [aTuple x] && tuple[1] == [aTuple y] && tuple[2] == [aTuple z] && tuple[3] == [aTuple w]) {
      return YES;
    }
  }
  return NO;
}

//-----------------------------------------------------------------------------
// Accessing
//-----------------------------------------------------------------------------

- (double *)elements
{
  return tuple;
}

- (void)setElements:(double *)values
{
  G3DCopyVector4dv(tuple,values);
}

- (void)getElements:(double *)values
{
  G3DCopyVector4dv(values,tuple);
}

- (double)x
{
  return tuple[0];
}

- (void)setX:(double)x
{
  tuple[0] = x;
}

- (double)y
{
  return tuple[1];
}

- (void)setY:(double)y
{
  tuple[1] = y;
}

- (double)z
{
  return tuple[2];
}

- (void)setZ:(double)z
{
  tuple[2] = z;
}

- (double)w
{
  return tuple[3];
}

- (void)setW:(double)w
{
  tuple[3] = w;
}

- (void)setValuesWithTuple:(G3DTuple4d *)aTuple
{
  double *tmp = [aTuple elements];
 
  tuple[0] = tmp[0];
  tuple[1] = tmp[1];
  tuple[2] = tmp[2];
  tuple[3] = tmp[3];
}

- (NSString *)description
{
  return [NSStringClass stringWithFormat:@"<%@ %x %x>: x = %f, y = %f, z = %f, w = %f",[self class],self,[NSThread currentThread],tuple[0],tuple[1],tuple[2],tuple[3]];
}

//-----------------------------------------------------------------------------
// NSCoding
//-----------------------------------------------------------------------------

- (void)encodeWithCoder:(NSCoder *)aCoder
{
  [aCoder encodeArrayOfObjCType:@encode(double) count:4 at:tuple];
}

- (id)initWithCoder:(NSCoder *)aCoder
{
  self = [super init];
  [aCoder decodeArrayOfObjCType:@encode(double) count:4 at:tuple];
  return self;
}

//-----------------------------------------------------------------------------
// NSCopying
//-----------------------------------------------------------------------------

- (id)copyWithZone:(NSZone *)zone
{
  return [[[self class] allocWithZone:zone] initWithTuple:self];
}

@end






